import cv2
import math
from gpiozero import LED
led = LED(17)

def calculate_distance(point1, point2):
    return math.sqrt((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2)

def detect_palm(hand_contour):
    # Find the bounding rectangle of the hand contour
    x, y, w, h = cv2.boundingRect(hand_contour)

    # Calculate the center of the bounding rectangle
    center_x = x + w // 2
    center_y = y + h // 2

    # Calculate the width and height of the palm region
    palm_width = int(0.9 * w)
    palm_height = int(0.6 * h)

    # Calculate the top-left corner of the palm region
    palm_x = center_x - palm_width // 2
    palm_y = center_y - palm_height // 2

    return palm_x, palm_y, palm_width, palm_height

def main():
    cap = cv2.VideoCapture(0)  # Open the webcam (change the index if using a different camera)

    while True:
        ret, frame = cap.read()
        if not ret:
            break

        # Convert the frame to grayscale
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        # Apply a Gaussian blur to reduce noise
        blurred = cv2.GaussianBlur(gray, (7, 7), 0)

        # Perform edge detection
        edges = cv2.Canny(blurred, 50, 150)

        # Find contours of the hand
        contours, _ = cv2.findContours(edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        if contours:
            # Find the hand contour with the largest area
            hand_contour = max(contours, key=cv2.contourArea)

            # Detect the palm region
            palm_x, palm_y, palm_width, palm_height = detect_palm(hand_contour)

            # Draw the palm rectangle
            cv2.rectangle(frame, (palm_x, palm_y), (palm_x + palm_width, palm_y + palm_height), (0, 255, 0), 2)

            # Find the defects in the hand contour
            defects = cv2.convexityDefects(hand_contour, cv2.convexHull(hand_contour, returnPoints=False))

            if defects is not None:
                finger_tips = []

                for i in range(defects.shape[0]):
                    s, e, f, _ = defects[i, 0]
                    start = tuple(hand_contour[s][0])
                    end = tuple(hand_contour[e][0])
                    far = tuple(hand_contour[f][0])

                    # Draw the finger skeleton lines within the palm region
                    if palm_x <= start[0] <= palm_x + palm_width and palm_x <= end[0] <= palm_x + palm_width:
                        cv2.line(frame, start, end, (255, 0, 0), 2)
                        cv2.circle(frame, far, 5, (0, 255, 0), -1)

                        # Store the finger tip coordinates
                        finger_tips.append(far)

                if len(finger_tips) >= 2:
                    # Calculate the distance between the thumb and index finger tips
                    distance = calculate_distance(finger_tips[0], finger_tips[1])

                    # Draw the distance text
                    cv2.putText(frame, f"Distance: {distance}", (50, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)

                    # Check if any finger touches the thumb
                    if distance < 30:
                        cv2.putText(frame, "Lights On", (50, 100), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)
                        led.on()

                    else:
                        cv2.putText(frame, "Lights Off", (50, 100), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)
                        led.off()

        cv2.imshow("Hand Tracking", frame)

        if cv2.waitKey(1) & 0xFF == ord('q'):
            break

    cap.release()
    cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
